/*
 * Copyright (C) 2012 Giorgio Vazzana
 *
 * This file is part of parole-conference.
 *
 * parole-conference is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * parole-conference is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <poll.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <opus.h>
#include <ncurses.h>

#include "audio.h"
#include "rw.h"
#include "notes.h"

#define VERSION "0.10alpha1"
#define DEBUG    0
#define LOGFILE "error.log"

#define MAX_NUMBER_OF_NODES 10
#define NICKLEN             16
#define TIMEOUT             250

/* 2ch, 20ms at 48kHz */
#define PCM20msLEN          (2*960)
/* 2ch, 10*20ms at 48kHz */
#define PCMBUFLEN           (10*PCM20msLEN)

#define UDP_PACKET_TYPE_CALL    0x0000
#define UDP_PACKET_TYPE_ANSWER  0x0001
#define UDP_PACKET_TYPE_TABLE   0x0002
#define UDP_PACKET_TYPE_AUDIO   0x0003
/* An audio udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   sequence_number (uint32, be)
 * Payload:
 *   data (4000 byte max)
 *
 * The field sequence_number must start at 0 for the first packet and increase
 * monotonically. The last packet must have sequence_number = UDP_PACKET_AUDIO_EOS_MARKER.
 * The payload data is the opus packet coming out of the encoder.
 */
#define UDP_PACKET_TYPE_CHAT    0x0004

#define UDP_PACKET_AUDIO_HEADER_LEN  6
#define UDP_PACKET_AUDIO_PAYLOAD_LEN 4000
#define UDP_PACKET_AUDIO_EOS_MARKER  0xFFFFFFFF
#define UDP_PACKET_MAXLEN            (UDP_PACKET_AUDIO_HEADER_LEN + UDP_PACKET_AUDIO_PAYLOAD_LEN)

struct udp_packet_st {
	struct sockaddr_in   addr;
	uint8_t              data[UDP_PACKET_MAXLEN];
	size_t               len;
};

struct node {
	int                  present;
	int                  knows_about_me;

	struct sockaddr_in   addr;
	char                 nick[NICKLEN];

	int                  timeout;
	uint32_t             sequence_number;
	uint32_t             last_sequence_number;

	/* opus decoder */
	OpusDecoder         *dec;
	opus_int16           pcm[PCMBUFLEN];
	opus_int32           pcmlen;
	int                  pcmfull;
	int                  packets_count;
	int                  packets_lost;
};

/* msgbuf size */
#define MBS 256

struct PContext {
	/* my nickname */
	char                 nick[NICKLEN];

	/* file descriptors */
	int                  sockfd;
	int                  stdinfd;

	int                  listen;
	int                  call;
	int                  autoaccept;

	char                 msgbuf[MBS];
	FILE                *fplog;

	/* ncurses */
	int                  rows, cols;
	WINDOW              *win;
	/* keyboard input */
	char                 linebuf[256];
	int                  linepos;

	int                  packets_lost_g;
	int                  packets_count_g;

	struct udp_packet_st saved_packet;
	int                  need_answer;
	char                *answer;

	/* array of nodes */
	struct node         *nodes;

	opus_int16           pcm[PCM20msLEN];
	opus_int32           pcmlen;
};

const unsigned int rate = 48000, channels = 2;

static volatile sig_atomic_t keep_going;
static volatile sig_atomic_t exit_child;

void signal_handler(int sig)
{
	if (sig == SIGINT)
		keep_going = 0;
	if (sig == SIGUSR1)
		exit_child = 1;
}

void die(struct PContext *pctx, int exit_code)
{
	/* shut down ncurses */
	endwin();

	if (pctx->msgbuf[0] != '\0') {
		fprintf(pctx->fplog, "%s", pctx->msgbuf);
		fprintf(stderr,      "%s", pctx->msgbuf);
	} else {
		fprintf(stderr, "An error occured, please check file '%s' for more info.", LOGFILE);
	}
	exit(exit_code);
}

void clear_input_line(struct PContext *pctx)
{
	pctx->linepos = 0;
	move(pctx->rows - 1, 0);
	clrtoeol();
	refresh();
}

void print_prompt(struct PContext *pctx)
{
	printw("%s> ", pctx->nick);
	refresh();
}

void print_question(struct PContext *pctx)
{
	clear_input_line(pctx);
	printw("? %s", pctx->msgbuf);
	refresh();
}

void log_msg(struct PContext *pctx)
{
	if (pctx->msgbuf[0] != '\0') {
		fprintf(pctx->fplog, "%s", pctx->msgbuf);
//FIXME
		fflush(pctx->fplog);
	}
}

void print_msg(struct PContext *pctx)
{
	int y, x;

	log_msg(pctx);

	getyx(stdscr, y, x);
	wprintw(pctx->win, "%s", pctx->msgbuf);
	wrefresh(pctx->win);
	move(y, x);
	refresh();
}


void check_opus_error_or_die(struct PContext *pctx, int error)
{
	if (error != OPUS_OK) {
		snprintf(pctx->msgbuf, MBS, "opus error: %s\n", opus_strerror(error));
		die(pctx, 1);
	}
}

opus_int32 convert_opus_bw(opus_int32 bandwidth)
{
	return bandwidth == OPUS_BANDWIDTH_NARROWBAND    ? 4000 :
	      (bandwidth == OPUS_BANDWIDTH_MEDIUMBAND    ? 6000 :
	      (bandwidth == OPUS_BANDWIDTH_WIDEBAND      ? 8000 :
	      (bandwidth == OPUS_BANDWIDTH_SUPERWIDEBAND ? 12000 :
	      (bandwidth == OPUS_BANDWIDTH_FULLBAND      ? 20000 : 0 /* AUTO */))));
}


int resolve_address(struct PContext *pctx, char *host, int port, struct sockaddr_in *address)
{
	struct hostent *hostinfo = NULL;
	char **ppchar;

#define PRINT snprintf(pctx->msgbuf, MBS,

	/* let's see if we already have a valid ip adress */
	if (inet_aton(host, &address->sin_addr)) {
		PRINT "Host: %s\n", host);
		print_msg(pctx);
	} else {
		/* get host info */
		hostinfo = gethostbyname(host);
		if (!hostinfo) {
			snprintf(pctx->msgbuf, MBS, "Cannot get info for host: %s\n", host);
			return -1;
		}

		/* print host info */
		PRINT "Host: %s\n", host);
		print_msg(pctx);
		PRINT "  Name: %s\n", hostinfo->h_name);
		print_msg(pctx);
/*		PRINT "  Aliases:");
		print_msg(pctx);
		for (ppchar = hostinfo->h_aliases; *ppchar; ppchar++) {
			PRINT " %s", *ppchar);
			print_msg(pctx);
		}
		PRINT "\n");
		print_msg(pctx);
		PRINT "  Type: %s\n", hostinfo->h_addrtype == AF_INET ? "AF_INET" : "AF_INET6");
		print_msg(pctx);*/
		PRINT "  Address list:");
		print_msg(pctx);
		for (ppchar = hostinfo->h_addr_list; *ppchar; ppchar++) {
			PRINT " %s", inet_ntoa(*(struct in_addr *)*ppchar) );
			print_msg(pctx);
		}
		PRINT "\n");
		print_msg(pctx);

		if (hostinfo->h_addrtype == AF_INET6) {
			snprintf(pctx->msgbuf, MBS, "AF_INET6 not supported.\n");
			return -1;
		}
	}

	PRINT "Port: %d\n\n", port);
	print_msg(pctx);

	/* fill in address */
	address->sin_family   = AF_INET;
	if (hostinfo)
		address->sin_addr = *(struct in_addr *) *hostinfo->h_addr_list;
	address->sin_port     = htons(port);

	return 0;
#undef PRINT
}

void send_udp_packet(struct PContext *pctx, struct udp_packet_st *udp_packet)
{
	ssize_t nw;

	nw = sendto(pctx->sockfd, udp_packet->data, udp_packet->len, 0,
		        (struct sockaddr *)&udp_packet->addr, sizeof(udp_packet->addr));
	if (nw == -1) {
		snprintf(pctx->msgbuf, MBS, "sendto(): %s\n", strerror(errno));
		die(pctx, 1);
	}
	if (nw < udp_packet->len) {
		snprintf(pctx->msgbuf, MBS, "sendto() returned %zd, expected %zd\n", nw, udp_packet->len);
		log_msg(pctx);
	}
}

void receive_udp_packet(struct PContext *pctx, struct udp_packet_st *udp_packet)
{
	ssize_t   nr;
	socklen_t addr_len;

	udp_packet->len = UDP_PACKET_MAXLEN;
	addr_len = sizeof(udp_packet->addr);
	nr = recvfrom(pctx->sockfd, udp_packet->data, udp_packet->len, 0,
		          (struct sockaddr *)&udp_packet->addr, &addr_len);
	if (nr == -1) {
		snprintf(pctx->msgbuf, MBS, "recvfrom(): %s\n", strerror(errno));
		die(pctx, 1);
	}

	udp_packet->len = nr;
}

int is_same_addr(struct sockaddr_in a, struct sockaddr_in b)
{
	if (a.sin_addr.s_addr == b.sin_addr.s_addr && a.sin_port == b.sin_port)
		return 1;
	else
		return 0;
}

void node_update_stats(struct PContext *pctx)
{
	int i;

	pctx->packets_lost_g  = 0;
	pctx->packets_count_g = 0;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i].present == 0)
			continue;
		pctx->packets_lost_g  += pctx->nodes[i].packets_lost;
		pctx->packets_count_g += pctx->nodes[i].packets_count;
	}
}

int node_count(struct PContext *pctx)
{
	int i, n;

	n = 0;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i].present)
			n++;
	}
	return n;
}

int node_search(struct PContext *pctx, struct sockaddr_in addr)
{
	int i, idx;

	idx = -1;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i].present == 0)
			continue;
		if (is_same_addr(addr, pctx->nodes[i].addr)) {
			idx = i;
			break;
		}
	}
	return idx;
}

int node_empty_slot_index(struct PContext *pctx)
{
	int i;

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i].present == 0)
			return i;
	}
	snprintf(pctx->msgbuf, MBS, "Error, max node capacity reached (%d nodes).\n", MAX_NUMBER_OF_NODES);
	die(pctx, 1);
	return -1;
}

void node_delete(struct PContext *pctx, int idx, int timeout)
{
	snprintf(pctx->msgbuf, MBS, "Node %s:%-5hu (%s) left%s\n",
	        inet_ntoa(pctx->nodes[idx].addr.sin_addr), ntohs(pctx->nodes[idx].addr.sin_port),
		    pctx->nodes[idx].nick, timeout ? " (timeout)" : "");
	print_msg(pctx);

	pctx->nodes[idx].present        = 0;
	pctx->nodes[idx].knows_about_me = 0;

	/* destroy decoder state */
	opus_decoder_destroy(pctx->nodes[idx].dec);
	pctx->nodes[idx].dec = NULL;
}

int node_add(struct PContext *pctx, struct sockaddr_in addr, const char *nick)
{
	int idx, error;

	idx = node_empty_slot_index(pctx);

	pctx->nodes[idx].present              = 1;
	pctx->nodes[idx].knows_about_me       = 0;
	pctx->nodes[idx].addr                 = addr;
	if (nick != NULL)
		memcpy(pctx->nodes[idx].nick, nick, NICKLEN);
	else
		memset(pctx->nodes[idx].nick,    0, NICKLEN);
	pctx->nodes[idx].timeout              = TIMEOUT;
	pctx->nodes[idx].sequence_number      = 0;
	pctx->nodes[idx].last_sequence_number = 0;

	/* create decoder state */
	pctx->nodes[idx].dec = opus_decoder_create(rate, channels, &error);
	check_opus_error_or_die(pctx, error);
	pctx->nodes[idx].pcmlen        = 0;
	pctx->nodes[idx].pcmfull       = 0;
	pctx->nodes[idx].packets_count = 0;
	pctx->nodes[idx].packets_lost  = 0;

	snprintf(pctx->msgbuf, MBS, "Node %s:%-5hu (%s) joined the network\n",
	        inet_ntoa(pctx->nodes[idx].addr.sin_addr), ntohs(pctx->nodes[idx].addr.sin_port),
	        pctx->nodes[idx].nick);
	print_msg(pctx);

	return idx;
}


void packet_call(struct PContext *pctx, struct udp_packet_st *udp_packet)
{
	udp_packet->len = 2+NICKLEN;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_CALL);
	memcpy(udp_packet->data+2, pctx->nick, NICKLEN);
}

void packet_answer_no(struct udp_packet_st *udp_packet)
{
	udp_packet->len = 2+2;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_ANSWER);
	memcpy(udp_packet->data+2, "NO", 2);
}

void packet_table(struct PContext *pctx, struct udp_packet_st *udp_packet, int exclude_idx)
{
	int i;
	uint8_t *d;
	size_t   l;
	uint16_t table_size;

	/* all nodes excluding the sender of the call request */
	table_size = node_count(pctx);

	/* nodes address table */
	d = udp_packet->data;
	l = 0;
	write_be16(d, UDP_PACKET_TYPE_TABLE); d+=2; l+=2;
	write_be16(d, table_size);            d+=2; l+=2;
	/* my nick */
	memcpy(d, pctx->nick, NICKLEN);       d+=NICKLEN; l+=NICKLEN;
	write_be32(d, 0);                     d+=4;  l+=4;
	write_be16(d, 0);                     d+=2;  l+=2;
	/* other nodes except packet sender */
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i].present == 0)
			continue;
		if (i == exclude_idx)
			continue;
		memcpy(d, pctx->nodes[i].nick, NICKLEN);                   d+=NICKLEN; l+=NICKLEN;
		write_be32(d, ntohl(pctx->nodes[i].addr.sin_addr.s_addr)); d+=4;  l+=4;
		write_be16(d, ntohs(pctx->nodes[i].addr.sin_port));        d+=2;  l+=2;
	}
	udp_packet->len = l;
}

void packet_chat(struct udp_packet_st *udp_packet, char *buf, size_t len)
{
	udp_packet->len = 2+len;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_CHAT);
	memcpy(udp_packet->data+2, buf, len);
}


struct address_book {
	int                n;
	char               nick[MAX_NUMBER_OF_NODES][NICKLEN];
	struct sockaddr_in addr[MAX_NUMBER_OF_NODES];
};

#define RES_ERROR -1
#define RES_OK 0
#define RES_NO 1
#define RES_NEED_ANSWER 2

int process_packet(struct PContext *pctx, struct udp_packet_st *udp_packet)
{
	int i, j, error, idx;
	uint8_t *d;
	size_t   l;
	uint16_t packet_type, table_size;
	struct address_book ab;
	struct udp_packet_st udp_packet_response;

	uint8_t    *opus_dec_packet;
	opus_int32  opus_dec_packetlen;
	opus_int32  pcmlen;

	d = udp_packet->data;
	l = udp_packet->len;

	if (udp_packet->len < 2)
		return RES_ERROR;

	packet_type = read_be16(d); d+=2; l-=2;

	/* do we know this node already? */
	idx = node_search(pctx, udp_packet->addr);
	/* reset timeout counter */
	if (idx != -1)
		pctx->nodes[idx].timeout = TIMEOUT;

	switch (packet_type) {
	case UDP_PACKET_TYPE_CALL:
		if (l != NICKLEN) {
			snprintf(pctx->msgbuf, MBS, "Invalid CALL packet received\n");
			log_msg(pctx);
			return RES_ERROR;
		}

		/* ask if we can accept the call */
		if (pctx->autoaccept == 0 && pctx->listen) {
			if (pctx->answer == NULL) {
				snprintf(pctx->msgbuf, MBS, "Call from:  %s:%-5hu  (%s)\n",
				         inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port), d);
				print_msg(pctx);
				snprintf(pctx->msgbuf, MBS, "Accept call? [y/n] ");
				print_question(pctx);
				pctx->saved_packet = *udp_packet;
				pctx->need_answer  = 1;
				return RES_NEED_ANSWER;
			} else {
				char ans = pctx->answer[0];

				pctx->need_answer = 0;
				pctx->answer      = NULL;

				clear_input_line(pctx);
				print_prompt(pctx);
				if (ans != 'y') {
					snprintf(pctx->msgbuf, MBS, "Call refused.\n");
					print_msg(pctx);
					udp_packet_response.addr = udp_packet->addr;
					packet_answer_no(&udp_packet_response);
					send_udp_packet(pctx, &udp_packet_response);
					return RES_NO;
				} else {
					snprintf(pctx->msgbuf, MBS, "Call accepted.\n");
					print_msg(pctx);
				}
			}
		}

		/* insert remote node */
		if (idx == -1) {
			idx = node_add(pctx, udp_packet->addr, (char *)d);
			pctx->nodes[idx].knows_about_me = 1;
		}

		/* send table */
		udp_packet_response.addr = udp_packet->addr;
		packet_table(pctx, &udp_packet_response, idx);
		send_udp_packet(pctx, &udp_packet_response);
		return RES_OK;


	case UDP_PACKET_TYPE_ANSWER:
		if (idx == -1) {
			snprintf(pctx->msgbuf, MBS, "Unexpected ANSWER packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			log_msg(pctx);
			return RES_ERROR;
		}
		if (l != 2 && l != 3) {
			snprintf(pctx->msgbuf, MBS, "Invalid ANSWER packet received\n");
			log_msg(pctx);
			return RES_ERROR;
		}
		if (strncmp((char *)d, "YES", 3) == 0)
			return RES_OK;
		else if (strncmp((char *)d, "NO", 2) == 0)
			return RES_NO;
		else
			return RES_ERROR;


	case UDP_PACKET_TYPE_TABLE:
		if (idx == -1) {
			snprintf(pctx->msgbuf, MBS, "Unexpected TABLE packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			log_msg(pctx);
			return RES_ERROR;
		}

		table_size = read_be16(d); d+=2; l-=2;
		if (l != (NICKLEN+6)*table_size) {
			snprintf(pctx->msgbuf, MBS, "Invalid TABLE packet received\n");
			log_msg(pctx);
			return RES_ERROR;
		}

		/* copy table to addres book */
		ab.n = table_size;
		for (i = 0; i < ab.n; i++) {
			memcpy(ab.nick[i], d, NICKLEN);                       d+=NICKLEN; l-=NICKLEN;
			if (i == 0) {
				ab.addr[i] = udp_packet->addr;                    d+=6;  l-=6;
			} else {
				ab.addr[i].sin_family      = AF_INET;
				ab.addr[i].sin_addr.s_addr = htonl(read_be32(d)); d+=4;  l-=4;
				ab.addr[i].sin_port        = htons(read_be16(d)); d+=2;  l-=2;
			}
		}

		/* packet sender knows about me */
		pctx->nodes[idx].knows_about_me = 1;
		memcpy(pctx->nodes[idx].nick, ab.nick[0], NICKLEN);

		/* insert unknown nodes */
		for (j = 0; j < ab.n; j++) {
			int tidx;

			tidx = node_search(pctx, ab.addr[j]);
			if (tidx == -1)
				tidx = node_add(pctx, ab.addr[j], ab.nick[j]);
		}

		/* send call request to those who don't know about me */
		for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
			if (pctx->nodes[i].present && pctx->nodes[i].knows_about_me == 0) {
				udp_packet_response.addr = pctx->nodes[i].addr;
				packet_call(pctx, &udp_packet_response);
				send_udp_packet(pctx, &udp_packet_response);
			}
		}
		return RES_OK;


	case UDP_PACKET_TYPE_AUDIO:
		if (idx == -1) {
#if DEBUG
			snprintf(pctx->msgbuf, MBS, "Unexpected AUDIO packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			log_msg(pctx);
#endif
			return RES_ERROR;
		}

		/* read packet header */
		pctx->nodes[idx].sequence_number = read_be32(d);
		d+=4; l-=4;

		/* ok, sequence numbers of packets must always increase */
		if (pctx->nodes[idx].sequence_number == 0 ||
		    pctx->nodes[idx].sequence_number > pctx->nodes[idx].last_sequence_number) {
			/* check if we lost packets */
			if (pctx->nodes[idx].sequence_number != pctx->nodes[idx].last_sequence_number + 1 &&
			    pctx->nodes[idx].packets_count != 0 &&
			    pctx->nodes[idx].sequence_number != 0 &&
			    pctx->nodes[idx].sequence_number != UDP_PACKET_AUDIO_EOS_MARKER) {
				error = pctx->nodes[idx].sequence_number - (pctx->nodes[idx].last_sequence_number + 1);
#if DEBUG
				snprintf(pctx->msgbuf, MBS, "Oops, %d audio packets lost.\n", error);
				log_msg(pctx);
#endif
				pctx->nodes[idx].packets_lost += error;
			}

			pctx->nodes[idx].last_sequence_number = pctx->nodes[idx].sequence_number;

			opus_dec_packet    = udp_packet->data + UDP_PACKET_AUDIO_HEADER_LEN;
			opus_dec_packetlen = udp_packet->len  - UDP_PACKET_AUDIO_HEADER_LEN;

			/* decode */
			if (PCMBUFLEN/channels - pctx->nodes[idx].pcmlen < 960) {
#if DEBUG
				snprintf(pctx->msgbuf, MBS, "Warning: packet from '%s', pcmbuf is full (%d)\n",
				        pctx->nodes[idx].nick, pctx->nodes[idx].pcmlen);
				log_msg(pctx);
#endif
				pctx->nodes[idx].pcmfull++;
				pctx->nodes[idx].pcmlen -= 960;
				memmove(pctx->nodes[idx].pcm, pctx->nodes[idx].pcm + channels * 960,
				       (PCMBUFLEN - channels * 960) * sizeof(pctx->nodes[idx].pcm[0]));
			}

			pcmlen = opus_decode(pctx->nodes[idx].dec, opus_dec_packet, opus_dec_packetlen,
				     pctx->nodes[idx].pcm + channels * pctx->nodes[idx].pcmlen,
				     PCMBUFLEN/channels - pctx->nodes[idx].pcmlen, 0);
			if (pcmlen < 0) {
				snprintf(pctx->msgbuf, MBS, "opus_decode() failed: %s\n", opus_strerror(pcmlen));
				die(pctx, 1);
			}
			if (pcmlen == 960) {
				pctx->nodes[idx].pcmlen += pcmlen;
			} else {
				snprintf(pctx->msgbuf, MBS, "Warning: pcmlen = %d, discarding pcm data\n", pcmlen);
				log_msg(pctx);
			}
			/* increment packets counter */
			pctx->nodes[idx].packets_count++;

			/* end-of-stream marker */
			if (pctx->nodes[idx].sequence_number == UDP_PACKET_AUDIO_EOS_MARKER)
				node_delete(pctx, idx, 0);
		} else {
#if DEBUG
			snprintf(pctx->msgbuf, MBS, "Oops, audio packet duplicated or out of order.\n");
			log_msg(pctx);
#endif
		}
		return RES_OK;


	case UDP_PACKET_TYPE_CHAT:
		if (idx == -1) {
			snprintf(pctx->msgbuf, MBS, "Unexpected CHAT packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			log_msg(pctx);
			return RES_ERROR;
		}

		snprintf(pctx->msgbuf, MBS, "%s> %s", pctx->nodes[idx].nick, d);
		wattron(pctx->win, A_BOLD);
		print_msg(pctx);
		wattroff(pctx->win, A_BOLD);
		return RES_OK;


	default:
		snprintf(pctx->msgbuf, MBS, "Unknown packet type\n");
		log_msg(pctx);
		break;
	}
	return RES_ERROR;
}

int mix_audio(struct PContext *pctx)
{
	int i, j, n;
	opus_int32 pcm[PCM20msLEN];
	opus_int32 v, peak;
	float gain;

	memset(pcm, 0, sizeof(pcm));

	n = 0;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i].present == 0)
			continue;

		if (pctx->nodes[i].pcmlen >= 960) {
			n++;

			for (j = 0; j < channels * 960; j++)
				pcm[j] += pctx->nodes[i].pcm[j];

			pctx->nodes[i].pcmlen -= 960;
			if (pctx->nodes[i].pcmlen)
				memmove(pctx->nodes[i].pcm, pctx->nodes[i].pcm + channels * 960,
				       (PCMBUFLEN - channels * 960) * sizeof(pctx->nodes[i].pcm[0]));
		}
	}

	if (n) {
		peak = 0;
		for (j = 0; j < channels * 960; j++) {
			v = abs(pcm[j]);
			if (v > peak)
				peak = v;
		}

		gain = (float)(INT16_MAX) / (float)peak;
		if (gain > 1.0f)
			gain = 1.0f;

		for (j = 0; j < channels * 960; j++)
			pctx->pcm[j] = lrintf((float)pcm[j] * gain);

		pctx->pcmlen = 960;
	} else {
		pctx->pcmlen = 0;
	}
	return n;
}
#if 0
pid_t play_ringtone(char *audiodevice, int verbose)
{
	pid_t pid;
	struct sigaction act;
	int ret;
	struct audio_data ad_playback;
#define D 500
	struct note  {
		double freq;
		int    duration; //in milliseconds
	} const song[] = {   //nokia ringtone
		{NOTE_E6,  D/2},
		{NOTE_D6,  D/2},
		{NOTE_F5d, D},
		{NOTE_G5d, D},
		{NOTE_C6d, D/2},
		{NOTE_B5,  D/2},
		{NOTE_D5,  D},
		{NOTE_E5,  D},
		{NOTE_B5,  D/2},
		{NOTE_A5,  D/2},
		{NOTE_C5d, D},
		{NOTE_E5,  D},
		{NOTE_A5,  2*D+D},
		{  0.0,    0}
	};
#undef D
	int n, i, samples, pos;
	int16_t *buf, *start;
	int      buflen;
	double duration = 0.0, v;

	pid = fork();
	if (pid == -1) {
		perror("fork()");
		exit(1);
	}

	if (pid == 0) {
		FILE *fp;

		fp = fopen("error1.log", "w");
		if (fp == NULL)
			exit(1);

		/* install signal handler */
		act.sa_handler = signal_handler;
		sigemptyset(&act.sa_mask);
		act.sa_flags = SA_RESTART;
		ret = sigaction(SIGUSR1, &act, NULL);
		if (ret == -1) {
			//fprintf(stderr, "Error: sigaction() failed.\n");
			exit(1);
		}

		/* set global variable */
		exit_child = 0;

		/* compute ringtone duration */
		for (n = 0; song[n].duration != 0; n++)
			duration += song[n].duration;
		duration /= 1000.0;
		duration += 2; /* +2s silence */

		/* allocate memory for the samples buffer */
		buflen = ceil(duration * rate);
		buf    = calloc(channels * sizeof(*buf), buflen);
		if (!buf) {
			//fprintf(stderr, "Error on calloc()\n");
			exit(1);
		}

		/* synthesize ringtone */
		pos = 0;
		for (n = 0; song[n].duration != 0; n++) {
			samples = rate * song[n].duration / 1000;
			for (i = 0; i < samples; i++) {
				v = 23000.0 * exp(-(double)i / (1.0 * samples)) * sin(2.0 * M_PI * song[n].freq * i / (double)rate);
				if (channels == 1) {
					buf[i + pos] = lrint(v);
				} else {
					buf[channels*i + pos    ] = lrint(v);
					buf[channels*i + pos + 1] = lrint(v);
				}
			}
			pos += channels*samples;
		}

		/* init audio */
		ret = audio_init(&ad_playback, audiodevice, SND_PCM_STREAM_PLAYBACK,
			             rate, channels, NULL, verbose >= 2 ? 1 : 0, fp);
		if (ret == -1)
			exit(1);

		/* play ringtone until interrupted */
		n = 0;
		while (1) {
			start   = buf + channels*n;
			samples = (buflen - n) > (rate/10) ? (rate/10) : (buflen - n);
			ret = audio_write(&ad_playback, start, samples, 0);
			if (ret < 0)
				exit(1);
			if (exit_child)
				break;
			n += rate/10;
			if (n >= buflen)
				n = 0;
		}

		/* clean up and exit */
		ret = audio_close(&ad_playback);
		free(buf);
		fclose(fp);
		exit(0);
	}

	return pid;
}
#endif

void usage()
{
	fprintf(stderr,
"Usage: parole-conference [options] [-l | -c host]\n"
"Options:\n"
"       -h             print help\n"
"       -v             increment verbosity level\n"
"       -n nick        set nickname\n"
"       -p port        listening port            (8110)\n"
"       -b bitrate     set bitrate in bit/s      (16000)\n"
"       -d audiodev    use audio device audiodev (hw:0,0)\n"
"       -a             autoaccept calls\n"
"Modes:\n"
"       -l             listen for calls\n"
"       -c host[:port] call host\n"
	);
}

int main(int argc, char *argv[])
{
	int                  ret, opt, i, error;
	int                  verbose = 0;
	struct sigaction     act;
	struct pollfd        pfds[6];
	int                  nfds, pollret;
	struct timeval       tea0, tea1; /* encode audio */
	struct timeval       tda0, tda1; /* decode audio */
	long int             t_enca, t_deca;
	int                  ic;

	/* Context */
	struct PContext      pctx;

	/* udp socket */
	int                  local_port = 8110;
	struct sockaddr_in   local_addr;
	char                *remote_host = NULL;
	int                  remote_port = 8110;
	struct sockaddr_in   remote_addr;
	struct udp_packet_st udp_packet;
	uint32_t             sequence_number;
	size_t               total_bytes_out;
	size_t               total_bytes_in;

	/* audio data */
	struct audio_data    ad_capture, ad_playback;
	char                *audiodevice = "hw:0,0";
	float                peak_percent;
	int                  keep_going_capture, overruns, underruns, mutemic;

	/* opus encoder */
	OpusEncoder         *enc;
	opus_int32           bitrate = 16000;
//	opus_int16           inbuffer[2*960];	/* 2ch, 20ms at 48kHz (FS * 0.02) */
	uint8_t              opus_enc_packet[4000];
	opus_int32           opus_enc_packetlen;


	srand(time(NULL));
	fprintf(stderr, "parole-conference, ver. %s\n\n", VERSION);

	memset(pctx.nick, 0, NICKLEN);
	pctx.listen      = 0;
	pctx.call        = 0;
	pctx.autoaccept  = 0;
	pctx.linebuf[0]  = '\0';
	pctx.linepos     = 0;


	/* parse command line options */
	while ((opt = getopt(argc, argv, "hvn:p:b:d:alc:")) != -1) {
		char *pc;

		switch (opt) {
		case 'h':
			usage();
			exit(0);
			break;
		case 'v':
			verbose++;
			break;
		case 'n':
			strncpy(pctx.nick, optarg, NICKLEN-1);
			pctx.nick[NICKLEN-1] = '\0';
			break;
		case 'p':
			local_port = atoi(optarg);
			break;
		case 'b':
			bitrate = atoi(optarg);
			break;
		case 'd':
			audiodevice = optarg;
			break;
		case 'a':
			pctx.autoaccept = 1;
			break;
		case 'l':
			pctx.listen = 1;
			break;
		case 'c':
			pctx.call = 1;
			pc = strchr(optarg, ':');
			if (pc) {
				*pc = '\0';
				pc++;
				if (*pc)
					remote_port = atoi(pc);
			}
			remote_host = optarg;
			break;
		default:
			fprintf(stderr, "\n");
			usage();
			exit(1);
		}
	}
#if 0
	fprintf(stderr, "verbose=%d, local_port=%d, bitrate=%d, audiodev=%s, autoaccept=%d, encrypt=%d, "
	        "keyfile=%s, record=%d, listen=%d, call=%d, remote_host=%s, remote_port=%d\n",
	        verbose, local_port, bitrate, audiodevice, autoaccept, encrypt, keyfile==NULL ? "NULL" : keyfile,
	        record, listen, call, remote_host==NULL ? "NULL" : remote_host, remote_port);
#endif

	pctx.fplog = fopen(LOGFILE, "w");
	if (pctx.fplog == NULL) {
		fprintf(stderr, "Error: cannot open log file '%s'.\n", LOGFILE);
		exit(1);
	}

	if (!pctx.listen && !pctx.call) {
		fprintf(stderr, "Please select one option in {-l,-c}.\n\n");
		usage();
		exit(1);
	}

	if (pctx.listen && pctx.call) {
		fprintf(stderr, "Please select only one option in {-l,-c}.\n\n");
		usage();
		exit(1);
	}

	/* if no nick was specified, choose a random one */
	if (pctx.nick[0] == '\0') {
		for (i = 0; i < 3; i++)
			pctx.nick[i] = 97 + (rand() % 26);
	}

	/* allocate memory for the context */
	pctx.nodes = calloc(MAX_NUMBER_OF_NODES, sizeof(struct node));
	if (pctx.nodes == NULL) {
		fprintf(stderr, "Error: calloc() failed.\n");
		exit(1);
	}

	/* init ncurses */
	initscr();
	if (has_colors() == FALSE) {
		snprintf(pctx.msgbuf, MBS, "Your terminal does not support color\n");
		die(&pctx, 1);
	}
	start_color();
	assume_default_colors(COLOR_WHITE, COLOR_BLUE);
	init_pair(1, COLOR_BLACK,  COLOR_WHITE);
	init_pair(2, COLOR_GREEN,  COLOR_WHITE);
	init_pair(3, COLOR_YELLOW, COLOR_WHITE);
	init_pair(4, COLOR_RED,    COLOR_WHITE);
	cbreak();
	noecho();
	keypad(stdscr, TRUE);
	getmaxyx(stdscr, pctx.rows, pctx.cols);

	/* first line, title */
	move(0, 0);
	printw(" parole-conference, ver. %s", VERSION);
	move(0, 0);
	chgat(-1, A_NORMAL, 1, NULL);

	/* status line */
	move(pctx.rows - 2, 0);
	addch(' ');
	move(pctx.rows - 2, 0);
	chgat(-1, A_NORMAL, 1, NULL);

	pctx.win = newwin(pctx.rows - 4, pctx.cols, 2, 0);
	if (pctx.win == NULL) {
		snprintf(pctx.msgbuf, MBS, "Could not create ncurses window\n");
		die(&pctx, 1);
	}
	scrollok(pctx.win, TRUE);
	move(pctx.rows - 1, 0);
	print_prompt(&pctx);


	/* This loop will be executed only once if we are in "call" mode,
	   or until the user presses CTRL-C in "listen" mode */
	while (1) {
		/* set global variable */
		keep_going = 1;

		pctx.need_answer     = 0;
		pctx.answer          = NULL;
		pctx.pcmlen          = 0;

		/* create socket */
		pctx.sockfd = socket(PF_INET, SOCK_DGRAM, 0);
		if (pctx.sockfd == -1) {
			snprintf(pctx.msgbuf, MBS, "socket(): %s\n", strerror(errno));
			die(&pctx, 1);
		}
		/* name socket (bind it to local_port) */
		local_addr.sin_family      = AF_INET;
		local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		local_addr.sin_port        = htons(local_port);
		ret = bind(pctx.sockfd, (struct sockaddr *)&local_addr, sizeof(local_addr));
		if (ret == -1) {
			snprintf(pctx.msgbuf, MBS, "bind(): %s\n", strerror(errno));
			die(&pctx, 1);

		}
		/* get stdin fd */
		pctx.stdinfd = fileno(stdin);
		if (pctx.stdinfd == -1) {
			snprintf(pctx.msgbuf, MBS, "fileno(): %s\n", strerror(errno));
			die(&pctx, 1);
		}

		/* file descriptor for socket is at pfds[0] */
		pfds[0].fd     = pctx.sockfd;
		pfds[0].events = POLLIN;
		nfds           = 1;
		/* file descriptor for stdin  is at pfds[1] */
		pfds[1].fd     = pctx.stdinfd;
		pfds[1].events = POLLIN;
		nfds           = 2;

		/* Init alsa audio capture */
		ret = audio_init(&ad_capture, audiodevice, SND_PCM_STREAM_CAPTURE,
			             rate, channels, pfds+nfds, verbose >= 2 ? 1 : 0, pctx.fplog);
		if (ret == -1) {
			pctx.msgbuf[0] = '\0';
			die(&pctx, 1);
		}
		nfds += ad_capture.nfds;

		/* Init alsa audio playback */
		ret = audio_init(&ad_playback, audiodevice, SND_PCM_STREAM_PLAYBACK,
			             rate, channels, pfds+nfds, verbose >= 2 ? 1 : 0, pctx.fplog);
		if (ret == -1) {
			pctx.msgbuf[0] = '\0';
			die(&pctx, 1);
		}
		/* we won't poll this device */
		//nfds += ad_playback.nfds;


		/* create encoder state */
		snprintf(pctx.msgbuf, MBS, "libopus: %s\n", opus_get_version_string());
		log_msg(&pctx);
		enc = opus_encoder_create(rate, channels, OPUS_APPLICATION_VOIP, &error);
		check_opus_error_or_die(&pctx, error);
#if 0
		/* set maximum bandwidth */
		error = opus_encoder_ctl(enc, OPUS_SET_MAX_BANDWIDTH(OPUS_BANDWIDTH_WIDEBAND));
		check_opus_error_or_die(&pctx, error);
#endif
		/* set bitrate */
		error = opus_encoder_ctl(enc, OPUS_SET_BITRATE(bitrate));
		check_opus_error_or_die(&pctx, error);
		/* set signal type */
		error = opus_encoder_ctl(enc, OPUS_SET_SIGNAL(OPUS_SIGNAL_VOICE));
		check_opus_error_or_die(&pctx, error);


		snprintf(pctx.msgbuf, MBS, "Welcome! Enter /h for help.\n");
		print_msg(&pctx);

		/* delete nodes */
		for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
			if (pctx.nodes[i].present == 0)
				continue;
			node_delete(&pctx, i, 0);
		}

		/* make a call */
		if (pctx.call) {
			/* resolve remote address */
			snprintf(pctx.msgbuf, MBS, "Resolving address...\n");
			print_msg(&pctx);
			ret = resolve_address(&pctx, remote_host, remote_port, &remote_addr);
			if (ret == -1)
				die(&pctx, 1);

			/* send a call request */
			node_add(&pctx, remote_addr, NULL);
			udp_packet.addr = remote_addr;
			packet_call(&pctx, &udp_packet);
			send_udp_packet(&pctx, &udp_packet);

			/* receive packet from remote host */
			snprintf(pctx.msgbuf, MBS, "Call request sent, waiting for reply...\n");
			print_msg(&pctx);
			receive_udp_packet(&pctx, &udp_packet);

			/* check if the other side accepted the call */
			ret = process_packet(&pctx, &udp_packet);
			if (ret == RES_ERROR || ret == RES_NO) {
				snprintf(pctx.msgbuf, MBS, "Call refused.\n");
				die(&pctx, 1);
			}
			snprintf(pctx.msgbuf, MBS, "Call accepted!\n");
			print_msg(&pctx);
		}

		if (pctx.listen) {
			snprintf(pctx.msgbuf, MBS, "Listening for incoming calls on udp port %d...\n", local_port);
			print_msg(&pctx);
		}

#if 0
		/* wait for a call */
		if (pctx.listen) {
			while (1) {
				int ans, ch;

				/* listen for a call request (this call will block) */
				snprintf(pctx.msgbuf, MBS, "Listening for incoming calls on udp port %d...\n", local_port);
				print_msg(&pctx);
				receive_udp_packet(&pctx, &udp_packet);

				/* is it a valid CALL packet? */
				ans = process_packet(&pctx, &udp_packet);
				if (ans == RES_ERROR)
					continue;

				if (ans == RES_NEED_ANSWER) {
					pid_t child, retpid;

					child = play_ringtone(audiodevice, verbose);

					ch = getch();
					pctx.linebuf[0] = (ch == 'y' ? 'y' : 'n');
					pctx.linebuf[1] = '\0';
					pctx.answer = pctx.linebuf;
					ans = process_packet(&pctx, &pctx.saved_packet);

					ret = kill(child, SIGUSR1);
					if (ret == -1) {
						snprintf(pctx.msgbuf, MBS, "kill(): %s\n", strerror(errno));
						die(&pctx, 1);
					}
					retpid = wait(NULL);
					if (retpid == -1) {
						snprintf(pctx.msgbuf, MBS, "wait(): %s\n", strerror(errno));
						die(&pctx, 1);
					}
				}
				if (ans == RES_OK)
					break;
			}
		}
#endif

		/* install signal handler */
		act.sa_handler = signal_handler;
		sigemptyset(&act.sa_mask);
		act.sa_flags = SA_RESTART;
		ret = sigaction(SIGINT, &act, NULL);
		if (ret == -1) {
			snprintf(pctx.msgbuf, MBS, "Error: sigaction() failed.\n");
			die(&pctx, 1);
		}

		/* start audio capture */
		ret = audio_start(&ad_capture);
		if (ret == -1) {
			pctx.msgbuf[0] = '\0';
			die(&pctx, 1);
		}
		keep_going_capture = 1;

		/* reset counters */
		t_enca  = 0;
		t_deca  = 0;
		ic      = 0; /* capture frame counter */

		total_bytes_out      = 0;
		total_bytes_in       = 0;
		peak_percent         = 0.0f;
		overruns             = 0;
		underruns            = 0;
		mutemic              = 0;

		/* main loop */
		while (keep_going || keep_going_capture) {
			/* poll: is audio_from_capture or udp_socket ready? */
			pollret = poll(pfds, nfds, -1);

			/* poll can fail and return -1 when a signal arrives. errno will be EINTR */
			if (pollret == -1) {
				if (errno == EINTR) {
					continue;
				} else {
					snprintf(pctx.msgbuf, MBS, "poll(): %s\n", strerror(errno));
					die(&pctx, 1);
				}
			}

			/* timeout should never happen really... */
			if (pollret == 0) {
				snprintf(pctx.msgbuf, MBS, "poll timeout\n");
				log_msg(&pctx);
				continue;
			}


			/** UDP SOCKET **/
			if (pfds[0].revents & POLLERR) {
				snprintf(pctx.msgbuf, MBS, "Socket POLLERR\n");
				die(&pctx, 1);
			}

			/* process data from socket if ready */
			if (pfds[0].revents & POLLIN) {
				gettimeofday(&tda0, NULL);
				/* read udp_packet */
				receive_udp_packet(&pctx, &udp_packet);
				total_bytes_in += udp_packet.len;

				ret = process_packet(&pctx, &udp_packet);

				gettimeofday(&tda1, NULL);
				t_deca = (tda1.tv_sec - tda0.tv_sec) * 1000000L + tda1.tv_usec - tda0.tv_usec;
				continue;
			}


			/** STDIN **/
			if (pfds[1].revents & POLLIN) {
				int len, n, ch;
				opus_int32 bandwidth;

				if (pctx.need_answer) {
					ch = getch();
					pctx.linebuf[0] = (ch == 'y' ? 'y' : 'n');
					pctx.linebuf[1] = '\0';
					pctx.answer = pctx.linebuf;
					process_packet(&pctx, &pctx.saved_packet);
					continue;
				}

				/* read an entire line */
				ch = getch();
				if (ch == KEY_BACKSPACE) {
					int y, x;

					if (pctx.linepos) {
						pctx.linepos--;
						getyx(stdscr, y, x);
						move(y, x-1);
						addch(' ');
						move(y, x-1);
						refresh();
					}
					continue;
				}
				pctx.linebuf[pctx.linepos++] = ch;
				/* read 130 characters max */
				if (pctx.linepos == 130) {
					clear_input_line(&pctx);
					print_prompt(&pctx);
					continue;
				}
				if (ch != '\n') {
					addch(ch);
					refresh();
					continue;
				}
				pctx.linebuf[pctx.linepos++] = '\0';
				clear_input_line(&pctx);
				print_prompt(&pctx);

				len = strlen(pctx.linebuf);

				if (!pctx.need_answer && len > 1 && pctx.linebuf[0] != '/') {
					packet_chat(&udp_packet, pctx.linebuf, len+1);
					/* send udp packet */
					for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
						if (pctx.nodes[i].present == 0)
							continue;
						if (pctx.nodes[i].knows_about_me) {
							udp_packet.addr = pctx.nodes[i].addr;
							send_udp_packet(&pctx, &udp_packet);
							total_bytes_out += udp_packet.len;
						}
					}
					snprintf(pctx.msgbuf, MBS, "%s> %s", pctx.nick, pctx.linebuf);
					wattron(pctx.win, A_BOLD);
					print_msg(&pctx);
					wattroff(pctx.win, A_BOLD);
				}

				if (!pctx.need_answer && len > 2 && pctx.linebuf[0] == '/') {
					switch (pctx.linebuf[1]) {
					case 'h':
						snprintf(pctx.msgbuf, MBS,
"help:\n"
"  /h             help\n"
"  /i             print info\n"
"  /m             toggle mic mute\n"
"  /b bitrate     change bitrate\n"
"  /k n           kill node n\n"
"  /c host[:port] call host\n"
"  /q             quit\n"
);
						print_msg(&pctx);
						break;
					case 'i':
						snprintf(pctx.msgbuf, MBS, "info:\n");
						print_msg(&pctx);

						snprintf(pctx.msgbuf, MBS, "  capture:   rb[%3.0f%%]   overrun[%2d]  mute[%s]\n",
						         100.0f * ad_capture.avail/(float)ad_capture.buffer_size,
						         overruns, mutemic ? " on" : "off");
						print_msg(&pctx);
						snprintf(pctx.msgbuf, MBS, "  playback:  rb[%3.0f%%]  underrun[%2d]\n",
						         100.0f * ad_playback.avail/(float)ad_playback.buffer_size, underruns);
						print_msg(&pctx);


						/* get bitrate */
						error = opus_encoder_ctl(enc, OPUS_GET_BITRATE(&bitrate));
						check_opus_error_or_die(&pctx, error);
						/* get bandwidth */
						error = opus_encoder_ctl(enc, OPUS_GET_BANDWIDTH(&bandwidth));
						check_opus_error_or_die(&pctx, error);
						snprintf(pctx.msgbuf, MBS, "  stream:    bitrate[%5d bit/s]  bandwidth[%d Hz]\n",
						        bitrate, convert_opus_bw(bandwidth));
						print_msg(&pctx);

						snprintf(pctx.msgbuf, MBS, "  nodes: %d\n", node_count(&pctx));
						print_msg(&pctx);
						for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
							if (pctx.nodes[i].present == 0)
								continue;
							snprintf(pctx.msgbuf, MBS, "    %d: %s:%-5hu %12s  pl[%3d/%4d]  pcm[len:%4d full:%d]\n", i,
							        inet_ntoa(pctx.nodes[i].addr.sin_addr),
							        ntohs(pctx.nodes[i].addr.sin_port), pctx.nodes[i].nick,
							        pctx.nodes[i].packets_lost, pctx.nodes[i].packets_count,
							        pctx.nodes[i].pcmlen, pctx.nodes[i].pcmfull);
							print_msg(&pctx);
						}
						break;
					case 'm':
						mutemic = (mutemic == 0) ? 1 : 0;
						snprintf(pctx.msgbuf, MBS, "mute: %s\n", mutemic ? "on" : "off");
						print_msg(&pctx);
						break;
					case 'b':
						bitrate = atoi(pctx.linebuf+3);
						if (bitrate < 6000 || bitrate > 512000)
							bitrate = 16000;
						/* set bitrate */
						error = opus_encoder_ctl(enc, OPUS_SET_BITRATE(bitrate));
						check_opus_error_or_die(&pctx, error);
						/* get bitrate */
						error = opus_encoder_ctl(enc, OPUS_GET_BITRATE(&bitrate));
						check_opus_error_or_die(&pctx, error);
						snprintf(pctx.msgbuf, MBS, "bitrate: %d bit/s\n", bitrate);
						print_msg(&pctx);
						break;
					case 'k':
						if (len > 4) {
							n = atoi(pctx.linebuf+3);
							if (0 <= n && n < MAX_NUMBER_OF_NODES) {
								if (pctx.nodes[n].present)
									node_delete(&pctx, n, 0);
							}
						}
						break;
					case 'c':
						{
							char *pc;
							struct sockaddr_in addr;
							int idx, port = 8110;

							pctx.linebuf[len-1] = '\0';
							pc = strchr(pctx.linebuf+3, ':');
							if (pc) {
								*pc = '\0';
								pc++;
								if (*pc)
									port = atoi(pc);
							}
							ret = resolve_address(&pctx, pctx.linebuf+3, port, &addr);
							if (ret != -1) {
								idx = node_search(&pctx, addr);
								if (idx == -1)
									idx = node_add(&pctx, addr, NULL);
								udp_packet.addr = addr;
								packet_call(&pctx, &udp_packet);
								send_udp_packet(&pctx, &udp_packet);
							}
						}
						break;
					case 'q':
						keep_going = 0;
						break;
					default:
						snprintf(pctx.msgbuf, MBS, "  Unknown command\n");
						print_msg(&pctx);
						break;
					}
				}
			}


			/** AUDIO CAPTURE **/
			audio_poll_descriptors_revents(&ad_capture);

			if (ad_capture.revents & POLLERR) {
				if (!(ad_capture.revents & POLLIN)) {
					snprintf(pctx.msgbuf, MBS, "Audio Capture POLLERR\n");
					die(&pctx, 1);
				}
			}

			/* process captured audio if ready */
			if (ad_capture.revents & POLLIN && keep_going_capture) {
				gettimeofday(&tea0, NULL);

				/* read ad_capture.period_size frames of audio */
				ret = audio_read(&ad_capture, 1);
				if (ret < 0) {
					/* overrun, let's handle it and try again */
					overruns++;
					ret = audio_recover(&ad_capture, ret, 1);
					if (ret == 0) {
						ret = audio_start(&ad_capture);
						if (ret == -1) {
							pctx.msgbuf[0] = '\0';
							die(&pctx, 1);
						}
						ret = audio_read(&ad_capture, 0);
						if (ret < 0) {
							pctx.msgbuf[0] = '\0';
							die(&pctx, 1);
						}
					} else {
						snprintf(pctx.msgbuf, MBS, "audio_recover() on capture failed\n");
						die(&pctx, 1);
					}
				}

				/* make sure we have read an entire frame */
				if (ad_capture.frames != ad_capture.alsabuffersize) {
					snprintf(pctx.msgbuf, MBS, "Error, we only read %ld frames instead of %lu.\n",
						    ad_capture.frames, ad_capture.alsabuffersize);
					die(&pctx, 1);
				}

				/* increment timeout counter, exit after 5s without receiving packets */
				for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
					if (pctx.nodes[i].present && pctx.nodes[i].knows_about_me) {
						pctx.nodes[i].timeout--;
						if (pctx.nodes[i].timeout == 0)
							node_delete(&pctx, i, 1);
					}
				}

				/* time to exit? */
				if (pctx.call && node_count(&pctx) == 0) {
					snprintf(pctx.msgbuf, MBS, "Quitting...\n");
					print_msg(&pctx);
					keep_going = 0;
				}
				if (keep_going == 0)
					keep_going_capture = 0;

				/* if there are no nodes connected, there is nothing to do */
				if (node_count(&pctx) == 0) {
					ic = 0;
					continue;
				}

				/* silence mic if necessary */
				if (mutemic)
					memset(ad_capture.alsabuffer, 0,
					       ad_capture.alsabuffersize * ad_capture.channels * sizeof(*ad_capture.alsabuffer));

				/* find peak in captured audio */
				audio_find_peak(&ad_capture);
				if (peak_percent < ad_capture.peak_percent)
					peak_percent = ad_capture.peak_percent;

				/* encoding */
				opus_enc_packetlen = opus_encode(enc, ad_capture.alsabuffer, ad_capture.frames, opus_enc_packet, 4000);
				if (opus_enc_packetlen < 0) {
					snprintf(pctx.msgbuf, MBS, "opus_encode() failed: %s\n", opus_strerror(opus_enc_packetlen));
					die(&pctx, 1);
				}

				/* forge udp packet */
				if (keep_going_capture == 0) {
					sequence_number    = UDP_PACKET_AUDIO_EOS_MARKER;
				} else {
					sequence_number    = ic;
				}
				/* udp header */
				write_be16(udp_packet.data, UDP_PACKET_TYPE_AUDIO);
				write_be32(udp_packet.data+2, sequence_number); /* packet sequence number */
				/* the payload is simply the opus packet */
				memcpy(udp_packet.data + UDP_PACKET_AUDIO_HEADER_LEN, opus_enc_packet, opus_enc_packetlen);
				udp_packet.len = UDP_PACKET_AUDIO_HEADER_LEN + opus_enc_packetlen;

				/* send udp packet */
				for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
					if (pctx.nodes[i].present == 0)
						continue;
					if (pctx.nodes[i].knows_about_me) {
						udp_packet.addr = pctx.nodes[i].addr;
						send_udp_packet(&pctx, &udp_packet);
						total_bytes_out += udp_packet.len;
					}
				}


				/* mix audio */
				if (mix_audio(&pctx)) {
					/* playback decoded data */
					ret = audio_write(&ad_playback, pctx.pcm, pctx.pcmlen, 1);
					if (ret < 0) {
						/* underrun, let's handle it and try again */
						underruns++;
						ret = audio_recover(&ad_playback, ret, 1);
						if (ret == 0) {
							ret = audio_write(&ad_playback, pctx.pcm, pctx.pcmlen, 0);
							if (ret < 0) {
								pctx.msgbuf[0] = '\0';
								die(&pctx, 1);
							}
						} else {
							snprintf(pctx.msgbuf, MBS, "audio_recover() on playback failed\n");
							die(&pctx, 1);
						}
					}
				}

				gettimeofday(&tea1, NULL);
				t_enca = (tea1.tv_sec - tea0.tv_sec) * 1000000L + tea1.tv_usec - tea0.tv_usec;

				/* increment frame counter */
				ic++;

				/* status bar */
				if ((ic%50) == 0) {
					int y, x, peakcolor;

					/* read available frames */
					ret = audio_avail(&ad_capture, 1);
					if (ret < 0)
						ad_capture.avail = ad_capture.buffer_size;
					ret = audio_avail(&ad_playback, 1);
					if (ret < 0)
						ad_playback.avail = 0;

					node_update_stats(&pctx);
					getyx(stdscr, y, x);

					peakcolor = (peak_percent <= 50.0f) ? 2 : ((peak_percent <= 80.0f) ? 3 : 4);

					move(pctx.rows - 2, 0);
					attron(COLOR_PAIR(1));
					printw("\r%02d:%02d  peak[", ic/3000, (ic/50)%60);
					attron(COLOR_PAIR(peakcolor));
					if (mutemic)
						printw("  --  ");
					else
						printw("%5.1f%%", peak_percent);
					attron(COLOR_PAIR(1));
					printw("]  speed[o:%4.1f i:%4.1f kB/s]  pl[%d/%4.2f%%]  xrun[o:%d u:%d]",
						    total_bytes_out/1024.0, total_bytes_in/1024.0,
						    pctx.packets_lost_g, 100.0f*pctx.packets_lost_g/(float)pctx.packets_count_g,
					        overruns, underruns);
					attroff(COLOR_PAIR(1));
					chgat(-1, A_NORMAL, 1, NULL);

/*					printw("\r%02d:%02d [rb:%3.0f%% pk:%5.1f%%] [o:%4.1f i:%4.1f kB/s] [rb:%3.0f%% pl:%d/%4.2f%% ur:%d]",
						    ic/3000, (ic/50)%60,
						    100.0f * ad_capture.avail/(float)ad_capture.buffer_size, peak_percent,
						    total_bytes_out/1024.0, total_bytes_in/1024.0,
						    100.0f * ad_playback.avail/(float)ad_playback.buffer_size,
						    pctx.packets_lost_g, 100.0f*pctx.packets_lost_g/(float)pctx.packets_count_g, underruns);
*/
					move(y, x);
					refresh();

					peak_percent    = 0.0f;
					total_bytes_out = 0;
					total_bytes_in  = 0;
				}
			}
		}

		/* check if we still have stuff on the socket */
		while ((pollret = poll(pfds, 1, 250)) > 0) {
			if (pfds[0].revents & POLLIN)
				receive_udp_packet(&pctx, &udp_packet);
		}

		/* reset signal handler */
		act.sa_handler = SIG_DFL;
		sigemptyset(&act.sa_mask);
		act.sa_flags = SA_RESTART;
		ret = sigaction(SIGINT, &act, NULL);
		if (ret == -1) {
			snprintf(pctx.msgbuf, MBS, "Error: sigaction() failed.\n");
			die(&pctx, 1);
		}

		/* clean up for the next iteration */
		ret = close(pctx.sockfd);
		ret = audio_stop(&ad_capture);
		ret = audio_close(&ad_capture);
		ret = audio_stop(&ad_playback);
		ret = audio_close(&ad_playback);

		/* destroy encoder state */
		opus_encoder_destroy(enc);

		/* repeat only if we are listening for calls */
		if (pctx.call)
			break;

		if (node_count(&pctx) == 0)
			break;
	}

	endwin();
	return 0;
}
